/* FILE: grid.h           -*-Mode: c++-*-
 *
 * This file contains class declarations for micromagnetic grid classes.
 *
 */

#ifndef _GRID_H
#define _GRID_H
#include <string.h>
#include <stdio.h>

#include "nb.h"

#include "demag.h"
#include "magelt.h"
#include "maginit.h"
#include "mifio.h"
#include "vector.h"
#include "zeeman.h"

/* End includes */

#if (SYSTEM_TYPE == UNIX)
#define STOPWATCH_ON  // #define for routine timing
#endif

#define COOKIELENGTH 8     // Length of file identifier (magic cookie) for
                           // G2DFil files.

#define PERTURBATION_SIZE 0.04 // Magnitude (relative to 1) of random pertur-
    // bations to magnetization; Used in magnetization initialiation routines.
    // The arctan of this is the maximum perturbation angle.

// Grid (mesh) file magic cookie string
extern const char GridFileMagic02[8];

//////////////////////// Grid2D Declaration //////////////////////////
class Grid2D {
private:
  int randomizer_seed;  // Set from mif file. 0 => random seed.
  double xmax,ymax,zmax,cellsize; // Dimensions in meters
  int Nx,Nz;            // Mesh size
  double moment_sum;    // ==Nx*Nz if all spins are unit spins
  double energy;        // Mesh energy density
  double exch_energy,anis_energy,demag_energy,zeeman_energy;  // All these
  // energies set by CalculateEnergy()
  Nb_Array2D<MagElt>m;     // Magnetization vector array
  Nb_Array2D<MagElt>m0;
  Nb_Array2D<MagElt>m1;
  Nb_Array2D<MagElt>m2;
  MagEltAnisType magelt_anistype; // MagElt Anisotropy type
  ThreeVector **anisa,**anisb,**anisc;
  int hValid; // ==0 => h is invalid for current m
  ThreeVector** h;      // h field vector array
  ThreeVector **h0,**h1;  // Temporary h field vector arrays
  ThreeVector** hdemag; // H demag field vector array
  Zeeman *hApplied; // Reduced applied field, constant across grid
  /// NOTE: hApplied stores, works and returns data in the OOMMF
  ///  xy+z coord system, as opposed to the xz+y coords used in
  ///  Grid2D.  All accesses to this class that are used internally
  ///  should pass the result through to ConvertXyzToXzy for the
  ///  appropriate conversions.

  ThreeVector **torque; // Torques for updating m
  ThreeVector **torque0,**torque1;
  int torqueValid; // ==0 => torque is invalid for current m and h
  int energyValid; // ==0 => energy is invalid for current m and h
  double** ADemagCoef;   // Coefficient matrix for demag calculations
  double** CDemagCoef;   // Coefficient matrix for demag calculations
  double ADemag(int i,int k);  // Returns element A(i,j) of A demag matrix
  double CDemag(int i,int k);  // Returns element C(i,j) of C demag matrix
  double SelfDemag;      // Self-demag coefficient
  double exchange_stiffness,Ms,K1;  // Material constants.
  // K1 is a nominal value (J/m^3) that may vary between mesh points.

  // ODE parameters
  int step_total,reject_total,reject_energy_count,reject_position_count;
  /// Step rejection counts.  The first should not be larger than the
  /// sum of the rest. These should be reset to 0 whenever hUpdateCount is. 
  /// NOTE: step_total is the total number of steps *attempted*.
  double GyRatio;  // Gyromagnetic ratio
  BOOL DoPrecess;    // If true, enable precession component in Torque.
  double DampCoef;   // Damping coefficient
  // dm/dt = (-GyRatio.Ms)(mxh)-(DampCoef.GyRatio.Ms)(mx(mxh))
  //  This program actually solves RHS=(-1/DampCoef)(mxh)-mx(mxh)
  //  Convert StepSize (below) to actual time (in seconds) via
  //              time_step = StepSize/(DampCoef.GyRatio.Ms)
  //   cf. Grid2d::GetTimeStep()
  double StepSize;
  double StepSize0; // Stepsize on previous pass
  double NextStepSize; // Suggested stepsize for next pass
  double InitialStepSize;  // Very first stepsize
  void Allocate(int xcount,int zcount);
  void Deallocate(void);
  void Initialize(MIF &mif,MagEltAnisType new_magelt_anistype);
  ArgLine MagInitArgs;  // Initial magnetization specification
  int FindNeighbors(int nsize,Nb_Array2D<MagElt> &me,MagEltLink men[],
		    int i,int k);

  ThreeVector& CalculateDemag(int i,int k); // Demag. vector at (i,j)
  void CalculateEnergy(void);
  int hUpdateCount; // Count of calls to hUpdate
  void hUpdate(void); // Calculates h (=h_effective) from m
  void hFastUpdate(void); // Same as hUpdate, but does not recalculate
                          // hdemag.
  int ODEIterCount; // ODE Iteration count for current segment.  Reset by
                    // ResetODE(), incremented by StepODE;  Used by
                    // multistep methods to determine if there is sufficient
                    // past history on this segment.
  void CalculateTorque(Nb_Array2D<MagElt> &m_in,
		       ThreeVector** &h_in,
		       ThreeVector** &t_out);

  // Interface to external/internal demag calculation routines
  DF_FuncInfo* ExternalDemag; // Set "->calc" to NULL to use internal routine
  // DF_Calc routines make use of externals DF_Export_m WriteGrid2Dm and
  // DF_Import_h FillGrid2Dh;
  char **dfparams; // Demag field routine parameter list, only used
                    // by Grid2D::Grid2D(char *filename)
  void InternalDemagCalc(void); // Internal, brute-force routine
  int DoDemag;  // 0 => no demag.

  // Integration routines
  void StepEuler(double maxtorque,double &next_stepsize); // 1st order Euler
  void StepPredict2(double maxtorque,double &next_stepsize);
  /// 2nd order predictor
  void IncRungeKutta4(Nb_Array2D<MagElt> &m_in,
		      ThreeVector** &h_in,
		      double step,
		      Nb_Array2D<MagElt> &m_out);
  void StepRungeKutta4(double maxtorque,double &next_stepsize,
		       int forcestep=0);
  /// 4th order Runge-Kutta

  // Utility routines
  void ConvertXyzToXzy(ThreeVector &vec); // xyz := OOMMF coordinate system,
  void ConvertXzyToXyz(ThreeVector &vec); // xzy := 2D solver coords.
  /// Corresponding functions for the Vec3D class are member functions
  /// of that class.
#ifdef STOPWATCH_ON
  Nb_StopWatch sw_exch,sw_anis,sw_demag,sw_solve_total,sw_proc_total;
#endif // STOPWATCH_ON
  int InitMagnetization();
  double GetMaxTorque(); // Returns maximum modulus of "torque," which
  /// is defined to be (-1/DampCoef)*(mxh) - mx(mxh).
  double GetMxH(double torque);  // Converts "torque" back to |mxh|.

public:
  // Birth and death routines
  // Standard constructor routine
  Grid2D(MIF mif);
  ~Grid2D(void);
  void Reset();  // Calling routine is responsible for resetting hApplied

#ifndef UNIT_MAG
  void InitRectangle(Nb_DString &a);
  void InitEllipse(Nb_DString &a );
  //Zero's moments outside ellipse defined by Nx,z
  void Mask(Nb_DString &a);
  void InitOval(Nb_DString &a);
  void SetBoundaryMoments(double momentsize);
       // Sets all boundary magelt's moment magnitudes to momentsize
  void RoughenBoundaryMoments();
       // Multiplies each boundary moment (separately) by an amount
       // drawn from a uniform distribution on [0,1].
#endif

  // Processing routines
  void Perturb(double max_mag);  // Randomly perturb all m by <max_mag
  double StepODE(int& errorcode); // Do one iteration
  void MarkhInvalid(void) { hValid=torqueValid=0; }
  void ResetODE(void) { ODEIterCount=0; MarkhInvalid(); }
  void SetAppliedField(const Vec3D &Bfield,int fieldstep) {
    hApplied->SetNomFieldVec3D(Ms,Bfield,fieldstep);
    ResetODE();
  }

  // Information routines
  void PrintUsageTimes(FILE *fptr); // Dumps CPU, system & elapsed times.
  void GetStepStats(int& _step_total,int& _reject_total,
		    int& _reject_energy_count,
		    int& _reject_position_count) {
    _step_total=step_total;
    _reject_total=reject_total;
    _reject_energy_count=reject_energy_count;
    _reject_position_count=reject_position_count;
  }

  double GetCellSize(void) { return cellsize; }
  double GetMaxNeighborAngle(void); // Returns maximum angle (in radians)
  // between neighboring spin directions in the grid.
  double GetEnergyDensity(void); // Returns average energy density, in J/m^3
  void GetEnergyDensities(double& exch,double& anis,double& demag,
			  double& zeeman,double& total);
  /// Returns average energy densities, in J/m^3

  double GetMaxMxH(); // Returns maximum value of |m x h|.
  double GetStepSize(void) { return StepSize; }
  double GetTimeStep(void); // Returns size of last step, in seconds
  void GetDimens(int& xcount,int& zcount) { xcount=Nx; zcount=Nz; }
  int GethUpdateCount(void) { return hUpdateCount; }

  // File output routines
  void BinDumpm(CHAR *filename,CHAR *note=NULL); // Reduced magnetization m
                                                 // in VecFil format
  void BinDumph(CHAR *filename,CHAR *note=NULL); // Reduced field h
                                                 // in VecFil format
  void WriteOvf(CHAR type,const CHAR *datastyle,const CHAR *filename,
		const CHAR *realname=NULL,const CHAR *note=NULL);
  /// Writes a 4-byte binary ovf file, of either m (if type=='m') or
  /// h (if type=='h').  The datastyle import should be one of "binary4",
  /// "binary8" or "text", with a NULL pointer or empty string implying
  /// the default value "binary4".

  // The following routines present a xy+z coords interface, performing
  // the conversions to/from the internal xz+y coords as necessary.
  void GetNearestGridNode(double x,double y,int &i,int &j);
  void GetRealCoords(int i,int j,double &x,double &y,double &z);
  double GetMs() const { return Ms; }
  void GetAverageMagnetization(ThreeVector& AveMag);
  Vec3D GetAveMag();  // Same as GetAverageMagnetization(), but returns Vec3D.
#ifdef CONTRIB_SPT001
  double GetResistance(); // CONTRIBUTED by Stephen Thompson,
  /// <mbgegspt@afs.mcc.ac.uk> 15-June-1998.
#endif // CONTRIB_SPT001

  Vec3D GetNomBApplied(void) { return hApplied->GetNomBField(); }
  void Dump(CHAR *filename,CHAR *note=NULL); // Dump of grid data in ASCII

  // The next routine is intended for internal use only, and therefore
  // the output data is kept in the internal xz+y coords system.
  void GetMag(double** arr); // Fills arr with magnetizations.
};

DF_Export_m WriteGrid2Dm;
DF_Import_h FillGrid2Dh;

#endif // _GRID_H
